import pathlib
import tempfile
import typing

import h5py
import matplotlib.pyplot as plt
import numpy as np
import obspy
import obspy.core.utcdatetime
import scipy.signal
import scipy.integrate
from tqdm import tqdm

import salvus.namespace as sn
import salvus.flow.collections
import salvus.project


class ExternalData:
    class Shotgather:
        class Trace:
            def __init__(self, trace: obspy.Trace) -> None:
                """
                Base class for individual seismic traces (receiver).

                Args:
                    trace: Obspy trace created from a stream.
                """
                self.data = trace.data
                self.channel = trace.stats.station
                self.location = trace.stats.receiver_location

        def __init__(self, waveforms: obspy.Stream) -> None:
            """
            Base class for shots (sources).

            It is assumed that each shot possesses a single source location.

            Args:
                waveforms: Obspy waveforms from a stream.

            Raises:
                AttributeError: Waveforms must contain either the sampling rate
                    or delta.
            """
            # Check that the metadata for all of the imported receivers are the same
            self._validate_metadata(waveforms)

            # Add the receivers to the object
            self.traces = []
            for trace in waveforms:
                self.traces.append(self.Trace(trace))

            # Add some metadata to the shot
            self.starttime = trace.stats.starttime
            self.endtime = trace.stats.endtime

            # Add the sampling rate and the delta depending on which one the user provided
            if hasattr(trace.stats, "sampling_rate") and hasattr(
                trace.stats, "delta"
            ):
                self.sampling_rate = trace.stats.sampling_rate
                self.dt = trace.stats.delta
            elif hasattr(trace.stats, "sampling_rate") and not hasattr(
                trace.stats, "delta"
            ):
                self.sampling_rate = trace.stats.sampling_rate
                self.dt = 1 / trace.stats.sampling_rate
            elif hasattr(trace.stats, "delta") and not hasattr(
                trace.stats, "sampling_rate"
            ):
                self.dt = trace.stats.delta
                self.sampling_rate = 1 / trace.stats.delta
            else:
                raise AttributeError(
                    "Waveforms must contain one of `delta` or `sampling_rate`"
                )

            self.n_pts = trace.stats.npts
            self.n_traces = len(self.traces)
            self.location = trace.stats.source_location
            self.name = str(trace.stats.source_id)

            # Add time array; assumes that the start time is at zero
            self.t = np.linspace(
                0.0, self.endtime - self.starttime, self.n_pts
            )

        def _validate_metadata(self, waveforms: obspy.Stream) -> None:
            """
            Verify that the metadata for all individual traces are identical.

            Args:
                waveforms: Waveforms to validate.

            Raises:
                ValueError: Metadata between traces must be identical.
            """

            def _get_ref_metadata(trace: obspy.Trace) -> dict:
                """
                Packages all of the metadata to validate in a single dictionary.

                Args:
                    trace: Trace to extract the metadata from.

                Returns:
                    Dictionary containing the information to validate.
                """
                meta = {}
                meta["starttime"] = float(trace.stats.starttime)
                meta["endtime"] = float(trace.stats.endtime)
                meta["sampling_rate"] = float(trace.stats.sampling_rate)
                meta["delta_t"] = float(trace.stats.delta)
                meta["npts"] = int(trace.stats.npts)

                return meta

            ref_meta = _get_ref_metadata(waveforms[0])

            for trace in waveforms:
                if ref_meta != _get_ref_metadata(trace):
                    raise ValueError(
                        "Metadata for all traces is not identical"
                    )

        def get_event(
            self, f: typing.Optional[np.ndarray] = None
        ) -> salvus.flow.collections.event.Event:
            """
            Converts a shot to a Salvus event.

            Extracts the information for the source for the shot (including the source and receiver locatios) and creates a Salvus event.

            Args:
                f: Source magnitude; must be of the same dimension as the
                    source/receiver data.  If left unspecified, `f` will be set
                    to `np.array([0.0, -1.0])` or `np.array([0.0, 0.0, -1.0])`
                    depending on the spatial dimension of the problem. Defaults
                    to None.

            Raises:
                ValueError: Number of dimensions passed to `f` must match the
                    number of dimensions of the problem.

            Returns:
                Data formatted as a Salvus event object.
            """
            ndim = len(self.location)

            # Set the default force if not specified
            if f is not None:
                if len(f) != ndim:
                    raise ValueError(
                        f"Number of dimensions for `f` expected {ndim}; got {len(f)}"
                    )
            elif ndim == 2:
                f = np.array([0.0, -1.0])
            else:
                f = np.array([0.0, 0.0, -1.0])

            if len(self.location) == 2:
                src = sn.simple_config.source.cartesian.VectorPoint2D(
                    x=self.location[0],
                    y=self.location[1],
                    fx=f[0],
                    fy=f[1],
                )

                recs = []
                for trace in self.traces:
                    recs.append(
                        sn.simple_config.receiver.cartesian.Point2D(
                            x=trace.location[0],
                            y=trace.location[1],
                            network_code="XX",
                            station_code=f"{trace.channel}",
                            fields=["velocity"],
                        )
                    )
            else:
                src = sn.simple_config.source.cartesian.VectorPoint3D(
                    x=self.location[0],
                    y=self.location[1],
                    z=self.location[2],
                    fx=f[0],
                    fy=f[1],
                    fz=f[2],
                )

                recs = []
                for trace in self.traces:
                    recs.append(
                        sn.simple_config.receiver.cartesian.Point3D(
                            x=trace.location[0],
                            y=trace.location[1],
                            z=trace.location[2],
                            network_code="XX",
                            station_code=f"{trace.channel}",
                            fields=["velocity"],
                        )
                    )

            return sn.Event(
                event_name=f"{self.name}", sources=src, receivers=recs
            )

        def plot_trace(
            self,
            trace_nr: int,
            t_range: typing.Optional[list] = None,
            freq_range: typing.Optional[list] = None,
        ) -> None:
            """
            Plot a single trace from the shot.

            Plot the time series and the power spectrum for a single trace from
            a given shot.

            Args:
                trace_nr: Trace index within
                    `ExternalData.shots[...].traces[trace_nr]`.
                t_range: Time range to limit the time series between; does not
                    change the power spectrum. Defaults to None.
                freq_range: Frequency range to limit the power spectrum within;
                    does not change the time series. Defaults to None.
            """
            trace = self.traces[trace_nr]

            # Initialize the plot
            fig, ax = plt.subplots(nrows=1, ncols=2, figsize=(15, 3))
            fig.suptitle(f"Shot `{self.name}`; Trace `{trace.channel}`")

            # Create the subplot for the time series
            ax[0].plot(self.t, trace.data, "k-")
            ax[0].set_xlabel("Time [s]")
            ax[0].set_ylabel("Amplitude")
            ax[0].set_title("Time Series")

            if t_range is not None:
                ax[0].set_xlim(left=t_range[0], right=t_range[1])

            ylim = 1.1 * np.abs(trace.data).max()
            ax[0].set_ylim(bottom=-ylim, top=ylim)

            # Get the power spectrum of the signal and normalize it
            power_spectrum = list(
                scipy.signal.periodogram(trace.data, fs=self.sampling_rate)
            )
            power_spectrum[1] = (
                power_spectrum[1] - power_spectrum[1].min()
            ) / (power_spectrum[1].max() - power_spectrum[1].min())

            # Create the subplot for the power spectrum
            ax[1].plot(power_spectrum[0], power_spectrum[1], "k-")
            ax[1].set_xlabel("Frequency [Hz]")
            ax[1].set_ylabel("Normalized Power")
            ax[1].set_title("Power Spectrum")

            # Get the xlimit for the power spectrum at where 99.9% of the max
            # power is
            cumulative = scipy.integrate.cumulative_trapezoid(
                power_spectrum[1], power_spectrum[0]
            )
            cumulative = (cumulative - cumulative.min()) / (
                cumulative.max() - cumulative.min()
            )
            max_power_ind = np.where(cumulative > 0.999)[0].min()
            if freq_range is not None:
                ax[1].set_xlim(left=freq_range[0], right=freq_range[1])
            else:
                ax[1].set_xlim(
                    left=0.0, right=power_spectrum[0][max_power_ind]
                )

            plt.show()

    def __init__(self, streams: typing.Union[list, obspy.Stream]) -> None:
        """
        Initialize a set of waveforms given one or more `obspy.Stream` objects.

        Take one or more `obspy.Stream` objects and package them into an `ExternalData` object.

        Args:
            streams: One or more Obspy streams.

        Raises:
            TypeError: Raised if the `streams` are not either a list or an Obspy
                Stream.
        """
        self.shots = []

        # Add only a single Stream
        if type(streams) == obspy.Stream:
            self.shots.append(self.Shotgather(streams))
        # Add a list of Streams
        elif type(streams) == list:
            for stream in streams:
                self.shots.append(self.Shotgather(stream))
        else:
            raise TypeError(f"Unknown input of type {type(streams)}")

    @property
    def n_shots(self) -> int:
        """
        Returns the number of shots in the current `ExternalData` object.

        Returns:
            Number of shots.
        """
        return len(self.shots)

    def get_earliest_starttime(
        self, return_index: bool = False
    ) -> typing.Union[
        obspy.core.utcdatetime.UTCDateTime,
        typing.Tuple[obspy.core.utcdatetime.UTCDateTime, int],
    ]:
        """
        Get the start time of the earliest shot.

        Args:
            return_index: Whether to return a tuple indicating the shot index
                which possesses the earliest start time. Defaults to False.

        Returns:
            Time of the earliest shot.
        """
        min_index = 0
        min_starttime = self.shots[0].starttime

        # Crude search over all of the shot start times
        for i, shot in enumerate(self.shots):
            test_starttime = shot.starttime
            if test_starttime < min_starttime:
                min_starttime = test_starttime
                min_index = i

        if return_index:
            return (min_starttime, min_index)
        else:
            return min_starttime

    def get_event(
        self, shot_name: str, f: typing.Optional[np.ndarray] = None
    ) -> salvus.flow.collections.event.Event:
        """
        Converts a shot to a Salvus event.

        Alias for `ExternalData.Shotgather.get_event()`.

        Args:
            shot_name: Name of the shot.
            f: Source magnitude; see docstring for
                `ExternalData.Shotgather.get_event()` for details. Defaults to
                None.

        Raises:
            ValueError: Raised if the shot is not found.

        Returns:
            Salvus event object.
        """
        for shot in self.shots:
            if shot.name == shot_name:
                return shot.get_event(f=f)

        raise ValueError(f"Shot name {shot_name} not found")

    def get_events(
        self, f: typing.Optional[np.ndarray] = None
    ) -> salvus.flow.collections.event_collection.EventCollection:
        """
        Converts all shots in the `ExternalData` object to an event collection.

        Extracts the source and receiver information for all shots within the
        `ExternalData` object and returns it as a Salvus `EventCollection`.
        This assumes that all events possess the same `f`.  To vary the source
        magnitude for each event, use `ExternalData.get_event()` instead.

        Args:
            f: Source magnitude; see docstring for
                `ExternalData.Shotgather.get_event()` for details. Defaults to None.

        Returns:
            Salvus event collection.
        """
        events = []
        for shot in self.shots:
            events.append(shot.get_event(f=f))

        return sn.EventCollection(events=events)

    def plot_shots(self) -> None:
        """
        Plot a schematic representation of all source positions.

        This assumes that the sources are all placed in-line along the x-axis.
        """
        plt.figure(figsize=(15, 3))

        for i, shot in enumerate(self.shots):
            loc = shot.location[:2]
            plt.scatter(*loc, color="orange")
            plt.annotate(
                text=shot.name,
                xy=loc,
                xytext=(-20, {0: 30, 1: -30, 2: 60, 3: -60}[i % 4]),
                textcoords="offset points",
                arrowprops={"arrowstyle": "->"},
            )
            plt.title("Shots")
        plt.xlabel("x-Position")
        plt.yticks([])
        plt.show()

    def plot_receivers(self, shot_name) -> None:
        """
        Plots a schematic representation of the receivers for a specific shot.

        This assumes that the receivers are all placed in-line along the x-axis.

        Args:
            shot_name: Name of the shot to plot.

        Raises:
            ValueError: Raised if the shot instance does not exist.
        """
        plt.figure(figsize=(15, 3))

        ind = -1
        for i, shot in enumerate(self.shots):
            if shot_name == shot.name:
                ind = i
                break

        if ind == -1:
            raise ValueError(f"Shot of name `{shot_name}` not found")

        for i, trace in enumerate(self.shots[ind].traces):
            loc = trace.location[:2]
            plt.scatter(*loc, color="green")
            plt.annotate(
                text=i,
                xy=loc,
                xytext=(-20, {0: 30, 1: -30, 2: 60, 3: -60}[i % 4]),
                textcoords="offset points",
                arrowprops={"arrowstyle": "->"},
            )
            plt.title(f"Receivers for Shot `{shot_name}`")
        plt.xlabel("x-Position")
        plt.yticks([])
        plt.show()

    def add_to_project(self, project: salvus.project.project.Project) -> None:
        """
        Adds the data from the `ExternalData` object to the Project.

        Adds the following details to the project from the `ExternalData`
        object:

        - Event based on the source and receivers positions for each
          `Shotgather`
        - Receiver data attached to the given event.

        Args:
            project: Salvus Project to add the external data to.
        """
        # Iterate over each shot
        for shot in tqdm(self.shots, ncols=100, desc="Adding shots"):
            # Add the event for each source position to the project
            event = shot.get_event()
            project.events.add(event)

            receiver_names = [
                f"{r.network_code}.{r.station_code}.{r.location_code}".encode()
                for r in event.receivers
            ]
            receiver_coordinates = np.array(
                [r.location for r in event.receivers]
            )

            # Create a single array containing all of the receiver data
            receiver_data = np.zeros([shot.n_traces, 1, shot.n_pts])
            for i, trace in enumerate(shot.traces):
                receiver_data[i, 0, :] = trace.data

            # Create a temporary HDF5 file where we can dump the relevant
            # information into
            with tempfile.TemporaryDirectory() as tmpdir:
                output_filename = pathlib.Path(tmpdir) / (shot.name + ".h5")

                # Add the receiver information to the HDF5 file
                with h5py.File(output_filename, "w") as f:
                    f.create_group("point")
                    f["point"].attrs["sampling_rate_in_hertz"] = (
                        1.0 / shot.dt,
                    )
                    f["point"].attrs["start_time_in_seconds"] = (shot.t.min(),)
                    f["point"]["velocity"] = receiver_data
                    f["point"]["velocity"].attrs[
                        "PHYSICS"
                    ] = "elastic".encode()

                    f["coordinates_ELASTIC_point"] = receiver_coordinates
                    f["names_ELASTIC_point"] = receiver_names

                # Save the information from the temporary HDF5 file into the project
                project.waveforms.add_external(
                    data_name="survey_data",
                    event=event.event_name,
                    data_filename=output_filename,
                )
